import torch


class RoundSTE(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        return torch.round(x)

    @staticmethod
    def backward(ctx, grad_output):
        return grad_output


def round_ste():
    def inner(x: torch.Tensor):
        return RoundSTE.apply(x)

    return inner


class BernoulliSTE(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        return torch.bernoulli(x)

    @staticmethod
    def backward(ctx, grad_output):
        return grad_output


def bernoulli_ste():
    def inner(x: torch.Tensor):
        return BernoulliSTE.apply(x)

    return inner


class ProbabilisticSTE(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        ctx.save_for_backward(x)
        return x * torch.bernoulli(x)

    @staticmethod
    def backward(ctx, grad_output):
        x, = ctx.saved_tensors
        return grad_output * 2 * x


def probabilistic_ste():
    def inner(x: torch.Tensor):
        return ProbabilisticSTE.apply(x)

    return inner



import torch
import matplotlib.pyplot as plt


def line_graph(list_of_values, title: str, label=None) -> None:
	"""
	Uses a list of values (allows tensors) to make a line graph
	:param list_of_values:
	:param title:
	:param label:
	:return:
	"""
	# Check if it's a list of tensors
	if isinstance(list_of_values[0], torch.Tensor):
		T = len(list_of_values)
		N = list_of_values[0].numel()

		# stack into shape (T, …)
		data = torch.stack(list_of_values, dim=0)  # if scalars, shape == (T,)
		if data.dim() == 1:
			data = data.unsqueeze(1)  # now shape == (T,1)

		data = data.cpu().numpy()
		x = range(T)

		for neuron_idx in range(N):
			if label is not None:
				plt.plot(x, data[:, neuron_idx], label=f'{label[neuron_idx]}')
			else:
				plt.plot(x, data[:, neuron_idx], label=f'Neuron {neuron_idx}')

		plt.title(title)
		if N > 1:
			plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left')
	else:
		# Assume list of floats (or ints)
		x = range(len(list_of_values))
		plt.plot(x, list_of_values)
		plt.title(title)

	plt.tight_layout()
	plt.show()



from ._spike_train import spike_train
from ._line_graph import line_graph
from ._measurement_manager import MeasurementManager
from ._render_image import render_image
from ._distributions import distributions


from typing import TypedDict, Optional, Literal, Union, Dict, Any
import torch
from torch import nn
from ._tt_infrastructure import TTLayer
from .. import functional


class LIB(TTLayer):
    def __init__(
            self,
            num_neurons: int,
            beta: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            beta_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            learn_beta: bool = True,
            learn_threshold: bool = True,
            spike_fn=functional.sigmoid4x,
            quant_fn: Literal["round", "bernoulli", "probabilistic"] = "bernoulli",
    ):
        super().__init__(num_neurons, dim)

        self._initialize_state("mem")
        self._register_decay("beta", beta, beta_rank, learn_beta)

        self.spike_fn = spike_fn
        quant_fn_map = {
            "round": self.round_ste,
            "bernoulli": self.bernoulli_ste,
            "probabilistic": self.probabilistic_ste
        }
        self.quant_fn = quant_fn_map.get(quant_fn, self.bernoulli_ste)
        self._register_threshold("threshold", threshold, threshold_rank, learn_threshold)

    def forward(self, x):
        self._ensure_states(x)
        x = self._to_working_dim(x)

        mem = self._to_working_dim(self.mem)
        mem = mem * self.beta + x

        spike_prob = self.spike_fn(mem - self.threshold)
        spikes = self.quant_fn(spike_prob)

        mem = mem - spikes * self.threshold

        spikes = self._from_working_dim(spikes)
        self.mem = self._from_working_dim(mem)

        return spikes


class DLIB(TTLayer):
    def __init__(
            self,
            num_neurons: int,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
            learn_threshold: bool = True,
            spike_fn=functional.sigmoid4x,
            quant_fn: Literal["round", "bernoulli", "probabilistic"] = "bernoulli",
    ):
        super().__init__(num_neurons, dim)

        self._initialize_state("pos_mem")
        self._initialize_state("neg_mem")
        self._register_decay("pos_beta", pos_beta, pos_beta_rank, learn_pos_beta)
        self._register_decay("neg_beta", neg_beta, neg_beta_rank, learn_neg_beta)

        self.spike_fn = spike_fn
        quant_fn_map = {
            "round": self.round_ste,
            "bernoulli": self.bernoulli_ste,
            "probabilistic": self.probabilistic_ste
        }
        self.quant_fn = quant_fn_map.get(quant_fn, self.bernoulli_ste)

        self._register_threshold("threshold", threshold, threshold_rank, learn_threshold)

    def forward(self, x):
        self._ensure_states(x)
        x = self._to_working_dim(x)

        pos_mem = self._to_working_dim(self.pos_mem)
        neg_mem = self._to_working_dim(self.neg_mem)
        pos_mem = pos_mem * self.pos_beta + torch.where(x >= 0, x, 0.0)
        neg_mem = neg_mem * self.neg_beta + torch.where(x <= 0, x, 0.0)

        mem = pos_mem + neg_mem

        spike_prob = self.spike_fn(mem - self.threshold)
        spikes = self.quant_fn(spike_prob)

        pos_mem = pos_mem - spikes * self.threshold * 0.5
        neg_mem = neg_mem - spikes * self.threshold * 0.5

        spikes = self._from_working_dim(spikes)
        self.pos_mem = self._from_working_dim(pos_mem)
        self.neg_mem = self._from_working_dim(neg_mem)

        return spikes


class SLIB(TTLayer):
    def __init__(
            self,
            num_neurons: int,
            alpha: Union[float, torch.Tensor] = 0.5,
            beta: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            alpha_rank: Literal[0, 1] = 1,
            beta_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            learn_alpha: bool = True,
            learn_beta: bool = True,
            learn_threshold: bool = True,
            spike_fn=functional.sigmoid4x,
            quant_fn: Literal["round", "bernoulli", "probabilistic"] = "bernoulli",
    ):
        super().__init__(num_neurons, dim)

        self._initialize_state("syn")
        self._register_decay("alpha", alpha, alpha_rank, learn_alpha)

        self._initialize_state("mem")
        self._register_decay("beta", beta, beta_rank, learn_beta)

        self.spike_fn = spike_fn
        quant_fn_map = {
            "round": self.round_ste,
            "bernoulli": self.bernoulli_ste,
            "probabilistic": self.probabilistic_ste
        }
        self.quant_fn = quant_fn_map.get(quant_fn, self.bernoulli_ste)

        self._register_threshold("threshold", threshold, threshold_rank, learn_threshold)

    def forward(self, x):
        self._ensure_states(x)
        x = self._to_working_dim(x)

        syn = self._to_working_dim(x)
        syn = syn * self.alpha + x * (1 - self.alpha)

        mem = self._to_working_dim(self.mem)
        mem = mem * self.beta + syn
        spike_prob = self.spike_fn(mem - self.threshold)
        spikes = self.quant_fn(spike_prob)

        mem = mem - spikes * self.threshold

        spikes = self._from_working_dim(spikes)
        self.syn = self._from_working_dim(syn)
        self.mem = self._from_working_dim(mem)

        return spikes


class RLIB(TTLayer):
    def __init__(
            self,
            num_neurons: int,
            beta: Union[float, torch.Tensor] = 0.9,
            gamma: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            rec_weight: Union[float, torch.Tensor] = 0.0,
            dim: int = -1,
            beta_rank: Literal[0, 1] = 1,
            gamma_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            rec_weight_rank: Literal[0, 1] = 1,
            learn_beta: bool = True,
            learn_gamma: bool = True,
            learn_threshold: bool = True,
            learn_rec_weight: bool = True,
            spike_fn=functional.sigmoid4x,
            quant_fn: Literal["round", "bernoulli", "probabilistic"] = "bernoulli",
    ):
        super().__init__(num_neurons, dim)

        self._initialize_state("mem")
        self._register_decay("beta", beta, beta_rank, learn_beta)

        self._initialize_state("rec")
        self._initialize_state("prev_output")
        self._register_decay("gamma", gamma, gamma_rank, learn_gamma)

        self.spike_fn = spike_fn
        quant_fn_map = {
            "round": self.round_ste,
            "bernoulli": self.bernoulli_ste,
            "probabilistic": self.probabilistic_ste
        }
        self.quant_fn = quant_fn_map.get(quant_fn, self.bernoulli_ste)

        self._register_threshold("threshold", threshold, threshold_rank, learn_threshold)

        self._register_parameter("rec_weight", rec_weight, rec_weight_rank, learn_rec_weight)

    def forward(self, x):
        self._ensure_states(x)
        x = self._to_working_dim(x)

        rec = self._to_working_dim(self.rec)
        prev_output = self._to_working_dim(self.prev_output)
        rec = rec * self.gamma + prev_output * (1 - self.gamma)
        self.rec = self._from_working_dim(rec)

        mem_delta = rec * self.rec_weight + x

        mem = self._to_working_dim(self.mem)
        mem = mem * self.beta + mem_delta

        spike_prob = self.spike_fn(mem - self.threshold)
        spikes = self.quant_fn(spike_prob)

        mem = mem - spikes * self.threshold

        spikes = self._from_working_dim(spikes)
        self.mem = self._from_working_dim(mem)
        self.prev_output = spikes

        return spikes


class DSLIB(TTLayer):
    def __init__(
            self,
            num_neurons: int,
            pos_alpha: Union[float, torch.Tensor] = 0.5,
            neg_alpha: Union[float, torch.Tensor] = 0.5,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            pos_alpha_rank: Literal[0, 1] = 1,
            neg_alpha_rank: Literal[0, 1] = 1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            learn_pos_alpha: bool = True,
            learn_neg_alpha: bool = True,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
            learn_threshold: bool = True,
            spike_fn=functional.sigmoid4x,
            quant_fn: Literal["round", "bernoulli", "probabilistic"] = "bernoulli",
    ):
        super().__init__(num_neurons, dim)

        self._initialize_state("pos_syn")
        self._initialize_state("neg_syn")
        self._register_decay("pos_alpha", pos_alpha, pos_alpha_rank, learn_pos_alpha)
        self._register_decay("neg_alpha", neg_alpha, neg_alpha_rank, learn_neg_alpha)

        self._initialize_state("pos_mem")
        self._initialize_state("neg_mem")
        self._register_decay("pos_beta", pos_beta, pos_beta_rank, learn_pos_beta)
        self._register_decay("neg_beta", neg_beta, neg_beta_rank, learn_neg_beta)

        self.spike_fn = spike_fn
        quant_fn_map = {
            "round": self.round_ste,
            "bernoulli": self.bernoulli_ste,
            "probabilistic": self.probabilistic_ste
        }
        self.quant_fn = quant_fn_map.get(quant_fn, self.bernoulli_ste)

        self._register_threshold("threshold", threshold, threshold_rank, learn_threshold)

    def forward(self, x):
        self._ensure_states(x)
        x = self._to_working_dim(x)

        pos_syn = self._to_working_dim(self.pos_syn)
        neg_syn = self._to_working_dim(self.neg_syn)
        pos_syn = pos_syn * self.pos_alpha + torch.where(x >= 0, x, 0.0) * (1 - self.pos_alpha)
        neg_syn = neg_syn * self.neg_alpha + torch.where(x <= 0, x, 0.0) * (1 - self.neg_alpha)

        self.pos_syn = self._from_working_dim(pos_syn)
        self.neg_syn = self._from_working_dim(neg_syn)

        syn = pos_syn + neg_syn

        pos_mem = self._to_working_dim(self.pos_mem)
        neg_mem = self._to_working_dim(self.neg_mem)
        pos_mem = pos_mem * self.pos_beta + torch.where(syn >= 0, syn, 0.0)
        neg_mem = neg_mem * self.neg_beta + torch.where(syn <= 0, syn, 0.0)

        mem = pos_mem + neg_mem

        spike_prob = self.spike_fn(mem - self.threshold)
        spikes = self.quant_fn(spike_prob)

        pos_mem = pos_mem - spikes * self.threshold * 0.5
        neg_mem = neg_mem - spikes * self.threshold * 0.5

        spikes = self._from_working_dim(spikes)
        self.pos_mem = self._from_working_dim(pos_mem)
        self.neg_mem = self._from_working_dim(neg_mem)

        return spikes


class DRLIB(TTLayer):
    def __init__(
            self,
            num_neurons: int,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            pos_gamma: Union[float, torch.Tensor] = 0.9,
            neg_gamma: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            pos_rec_weight: Union[float, torch.Tensor] = 0.0,
            neg_rec_weight: Union[float, torch.Tensor] = 0.0,
            dim: int = -1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            pos_gamma_rank: Literal[0, 1] = 1,
            neg_gamma_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            pos_rec_weight_rank: Literal[0, 1] = 1,
            neg_rec_weight_rank: Literal[0, 1] = 1,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
            learn_pos_gamma: bool = True,
            learn_neg_gamma: bool = True,
            learn_threshold: bool = True,
            learn_pos_rec_weight: bool = True,
            learn_neg_rec_weight: bool = True,
            spike_fn=functional.sigmoid4x,
            quant_fn: Literal["round", "bernoulli", "probabilistic"] = "bernoulli",
    ):
        super().__init__(num_neurons, dim)

        self._initialize_state("pos_mem")
        self._initialize_state("neg_mem")
        self._register_decay("pos_beta", pos_beta, pos_beta_rank, learn_pos_beta)
        self._register_decay("neg_beta", neg_beta, neg_beta_rank, learn_neg_beta)

        self._initialize_state("pos_rec")
        self._initialize_state("neg_rec")
        self._initialize_state("prev_output")
        self._register_decay("pos_gamma", pos_gamma, pos_gamma_rank, learn_pos_gamma)
        self._register_decay("neg_gamma", neg_gamma, neg_gamma_rank, learn_neg_gamma)

        self.spike_fn = spike_fn
        quant_fn_map = {
            "round": self.round_ste,
            "bernoulli": self.bernoulli_ste,
            "probabilistic": self.probabilistic_ste
        }
        self.quant_fn = quant_fn_map.get(quant_fn, self.bernoulli_ste)

        self._register_threshold("threshold", threshold, threshold_rank, learn_threshold)

        self._register_parameter("pos_rec_weight", pos_rec_weight, pos_rec_weight_rank, learn_pos_rec_weight)
        self._register_parameter("neg_rec_weight", neg_rec_weight, neg_rec_weight_rank, learn_neg_rec_weight)

    def forward(self, x):
        self._ensure_states(x)
        x = self._to_working_dim(x)

        pos_rec = self._to_working_dim(self.pos_rec)
        neg_rec = self._to_working_dim(self.neg_rec)
        prev_output = self._to_working_dim(self.prev_output)

        pos_rec = pos_rec * self.pos_gamma + torch.where(prev_output >= 0, prev_output, 0.0) * (1 - self.pos_gamma)
        neg_rec = neg_rec * self.neg_gamma + torch.where(prev_output <= 0, prev_output, 0.0) * (1 - self.neg_gamma)

        self.pos_rec = self._from_working_dim(pos_rec)
        self.neg_rec = self._from_working_dim(neg_rec)

        rec = pos_rec + neg_rec

        pos_mem_delta = torch.where(rec >= 0, rec, 0.0) * self.pos_rec_weight + torch.where(x >= 0, x, 0.0)
        neg_mem_delta = torch.where(rec <= 0, rec, 0.0) * self.neg_rec_weight + torch.where(x <= 0, x, 0.0)

        pos_mem = self._to_working_dim(self.pos_mem)
        neg_mem = self._to_working_dim(self.neg_mem)
        pos_mem = pos_mem * self.pos_beta + pos_mem_delta
        neg_mem = neg_mem * self.neg_beta + neg_mem_delta

        mem = pos_mem + neg_mem

        spike_prob = self.spike_fn(mem - self.threshold)
        spikes = self.quant_fn(spike_prob)

        pos_mem = pos_mem - spikes * self.threshold * 0.5
        neg_mem = neg_mem - spikes * self.threshold * 0.5

        spikes = self._from_working_dim(spikes)
        self.pos_mem = self._from_working_dim(pos_mem)
        self.neg_mem = self._from_working_dim(neg_mem)
        self.prev_output = spikes

        return spikes


class SRLIB(TTLayer):
    def __init__(
            self,
            num_neurons: int,
            alpha: Union[float, torch.Tensor] = 0.5,
            beta: Union[float, torch.Tensor] = 0.9,
            gamma: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            rec_weight: Union[float, torch.Tensor] = 0.0,
            dim: int = -1,
            alpha_rank: Literal[0, 1] = 1,
            beta_rank: Literal[0, 1] = 1,
            gamma_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            rec_weight_rank: Literal[0, 1] = 1,
            learn_alpha: bool = True,
            learn_beta: bool = True,
            learn_gamma: bool = True,
            learn_threshold: bool = True,
            learn_rec_weight: bool = True,
            spike_fn=functional.sigmoid4x,
            quant_fn: Literal["round", "bernoulli", "probabilistic"] = "bernoulli",
    ):
        super().__init__(num_neurons, dim)

        self._initialize_state("syn")
        self._register_decay("alpha", alpha, alpha_rank, learn_alpha)

        self._initialize_state("mem")
        self._register_decay("beta", beta, beta_rank, learn_beta)

        self._initialize_state("rec")
        self._initialize_state("prev_output")
        self._register_decay("gamma", gamma, gamma_rank, learn_gamma)

        self.spike_fn = spike_fn
        quant_fn_map = {
            "round": self.round_ste,
            "bernoulli": self.bernoulli_ste,
            "probabilistic": self.probabilistic_ste
        }
        self.quant_fn = quant_fn_map.get(quant_fn, self.bernoulli_ste)

        self._register_threshold("threshold", threshold, threshold_rank, learn_threshold)

        self._register_parameter("rec_weight", rec_weight, rec_weight_rank, learn_rec_weight)

    def forward(self, x):
        self._ensure_states(x)
        x = self._to_working_dim(x)

        syn = self._to_working_dim(x)
        syn = syn * self.alpha + x * (1 - self.alpha)
        self.syn = self._from_working_dim(syn)

        rec = self._to_working_dim(self.rec)
        prev_output = self._to_working_dim(self.prev_output)
        rec = rec * self.gamma + prev_output * (1 - self.gamma)
        self.rec = self._from_working_dim(rec)

        mem_delta = rec * self.rec_weight + syn

        mem = self._to_working_dim(self.mem)
        mem = mem * self.beta + mem_delta
        spike_prob = self.spike_fn(mem - self.threshold)
        spikes = self.quant_fn(spike_prob)

        mem = mem - spikes * self.threshold

        spikes = self._from_working_dim(spikes)
        self.mem = self._from_working_dim(mem)
        self.prev_output = spikes

        return spikes


class DSRLIB(TTLayer):
    def __init__(
            self,
            num_neurons: int,
            pos_alpha: Union[float, torch.Tensor] = 0.5,
            neg_alpha: Union[float, torch.Tensor] = 0.5,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            pos_gamma: Union[float, torch.Tensor] = 0.9,
            neg_gamma: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            pos_rec_weight: Union[float, torch.Tensor] = 0.0,
            neg_rec_weight: Union[float, torch.Tensor] = 0.0,
            dim: int = -1,
            pos_alpha_rank: Literal[0, 1] = 1,
            neg_alpha_rank: Literal[0, 1] = 1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            pos_gamma_rank: Literal[0, 1] = 1,
            neg_gamma_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            pos_rec_weight_rank: Literal[0, 1] = 1,
            neg_rec_weight_rank: Literal[0, 1] = 1,
            learn_pos_alpha: bool = True,
            learn_neg_alpha: bool = True,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
            learn_pos_gamma: bool = True,
            learn_neg_gamma: bool = True,
            learn_threshold: bool = True,
            learn_pos_rec_weight: bool = True,
            learn_neg_rec_weight: bool = True,
            spike_fn=functional.sigmoid4x,
            quant_fn: Literal["round", "bernoulli", "probabilistic"] = "bernoulli",
    ):
        super().__init__(num_neurons, dim)

        self._initialize_state("pos_syn")
        self._initialize_state("neg_syn")
        self._register_decay("pos_alpha", pos_alpha, pos_alpha_rank, learn_pos_alpha)
        self._register_decay("neg_alpha", neg_alpha, neg_alpha_rank, learn_neg_alpha)

        self._initialize_state("pos_mem")
        self._initialize_state("neg_mem")
        self._register_decay("pos_beta", pos_beta, pos_beta_rank, learn_pos_beta)
        self._register_decay("neg_beta", neg_beta, neg_beta_rank, learn_neg_beta)

        self._initialize_state("pos_rec")
        self._initialize_state("neg_rec")
        self._initialize_state("prev_output")
        self._register_decay("pos_gamma", pos_gamma, pos_gamma_rank, learn_pos_gamma)
        self._register_decay("neg_gamma", neg_gamma, neg_gamma_rank, learn_neg_gamma)

        self.spike_fn = spike_fn
        quant_fn_map = {
            "round": self.round_ste,
            "bernoulli": self.bernoulli_ste,
            "probabilistic": self.probabilistic_ste
        }
        self.quant_fn = quant_fn_map.get(quant_fn, self.bernoulli_ste)

        self._register_threshold("threshold", threshold, threshold_rank, learn_threshold)

        self._register_parameter("pos_rec_weight", pos_rec_weight, pos_rec_weight_rank, learn_pos_rec_weight)
        self._register_parameter("neg_rec_weight", neg_rec_weight, neg_rec_weight_rank, learn_neg_rec_weight)

    def forward(self, x):
        self._ensure_states(x)
        x = self._to_working_dim(x)

        pos_syn = self._to_working_dim(self.pos_syn)
        neg_syn = self._to_working_dim(self.neg_syn)
        pos_syn = pos_syn * self.pos_alpha + torch.where(x >= 0, x, 0.0) * (1 - self.pos_alpha)
        neg_syn = neg_syn * self.neg_alpha + torch.where(x <= 0, x, 0.0) * (1 - self.neg_alpha)

        self.pos_syn = self._from_working_dim(pos_syn)
        self.neg_syn = self._from_working_dim(neg_syn)

        syn = pos_syn + neg_syn

        pos_rec = self._to_working_dim(self.pos_rec)
        neg_rec = self._to_working_dim(self.neg_rec)
        prev_output = self._to_working_dim(self.prev_output)

        pos_rec = pos_rec * self.pos_gamma + torch.where(prev_output >= 0, prev_output, 0.0) * (1 - self.pos_gamma)
        neg_rec = neg_rec * self.neg_gamma + torch.where(prev_output <= 0, prev_output, 0.0) * (1 - self.neg_gamma)

        self.pos_rec = self._from_working_dim(pos_rec)
        self.neg_rec = self._from_working_dim(neg_rec)

        rec = pos_rec + neg_rec

        pos_mem_delta = torch.where(rec >= 0, rec, 0.0) * self.pos_rec_weight + torch.where(syn >= 0, syn, 0.0)
        neg_mem_delta = torch.where(rec <= 0, rec, 0.0) * self.neg_rec_weight + torch.where(syn <= 0, syn, 0.0)

        pos_mem = self._to_working_dim(self.pos_mem)
        neg_mem = self._to_working_dim(self.neg_mem)
        pos_mem = pos_mem * self.pos_beta + pos_mem_delta
        neg_mem = neg_mem * self.neg_beta + neg_mem_delta

        mem = pos_mem + neg_mem

        spike_prob = self.spike_fn(mem - self.threshold)
        spikes = self.quant_fn(spike_prob)

        pos_mem = pos_mem - spikes * self.threshold * 0.5
        neg_mem = neg_mem - spikes * self.threshold * 0.5

        spikes = self._from_working_dim(spikes)
        self.pos_mem = self._from_working_dim(pos_mem)
        self.neg_mem = self._from_working_dim(neg_mem)
        self.prev_output = spikes

        return spikes



from ._tt_infrastructure import TTLayer, TTModel

from ._li_layers import LI, DLI, SLI, DSLI
from ._lib_layers import LIB, DLIB, SLIB, RLIB, DSLIB, DRLIB, SRLIB, DSRLIB
from ._lit_layers import LIT, DLIT, SLIT, RLIT, DSLIT, DRLIT, SRLIT, DSRLIT
from ._lits_layers import LITS, DLITS, SLITS, RLITS, DSLITS, DRLITS, SRLITS, DSRLITS

# from . import flex



from typing import TypedDict, Optional, Literal, Union, Dict, Any
from ._leaky_integrator import LeakyIntegrator
import torch
from torch import nn


class LI(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            beta: Union[float, torch.Tensor] = 0.9,
            dim: int = -1,
            beta_rank: Literal[0, 1] = 1,
            learn_beta: bool = True,
    ):
        beta_setup = {
            "value": beta,
            "rank": beta_rank,
            "ema": False,
            "learnable": learn_beta,
        }

        super().__init__(
            num_neurons=num_neurons,
            dim=dim,
            pos_beta_setup=beta_setup,
        )


class DLI(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            dim: int = -1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
    ):
        pos_beta_setup = {
            "value": pos_beta,
            "rank": pos_beta_rank,
            "ema": False,
            "learnable": learn_pos_beta,
        }

        neg_beta_setup = {
            "value": neg_beta,
            "rank": neg_beta_rank,
            "ema": False,
            "learnable": learn_neg_beta,
        }

        super().__init__(
            num_neurons=num_neurons,
            dim=dim,
            pos_beta_setup=pos_beta_setup,
            neg_beta_setup=neg_beta_setup,
        )


class SLI(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            alpha: Union[float, torch.Tensor] = 0.5,
            beta: Union[float, torch.Tensor] = 0.9,
            dim: int = -1,
            alpha_rank: Literal[0, 1] = 1,
            beta_rank: Literal[0, 1] = 1,
            learn_alpha: bool = True,
            learn_beta: bool = True,
    ):
        alpha_setup = {
            "value": alpha,
            "rank": alpha_rank,
            "ema": True,
            "learnable": learn_alpha,
        }

        beta_setup = {
            "value": beta,
            "rank": beta_rank,
            "ema": False,
            "learnable": learn_beta,
        }

        super().__init__(
            num_neurons=num_neurons,
            dim=dim,
            pos_alpha_setup=alpha_setup,
            pos_beta_setup=beta_setup,
        )


class DSLI(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            pos_alpha: Union[float, torch.Tensor] = 0.5,
            neg_alpha: Union[float, torch.Tensor] = 0.5,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            dim: int = -1,
            pos_alpha_rank: Literal[0, 1] = 1,
            neg_alpha_rank: Literal[0, 1] = 1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            learn_pos_alpha: bool = True,
            learn_neg_alpha: bool = True,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
    ):
        pass



import torch
from torch import nn
import tracetorch as tt
from tracetorch import snn

import pytest


class TestLayerEquivalence:
    """Test equivalence between base and flex layer implementations"""

    @pytest.fixture(autouse=True)
    def setup(self):

        torch.manual_seed(42)
        self.num_neurons = 10
        self.batch_size = 100
        self.num_timesteps = 100
        self.dim = -1
        self.lr = 1e-2

        self.pos_alpha = torch.rand(self.num_neurons)
        self.neg_alpha = torch.rand(self.num_neurons)
        self.pos_beta = torch.rand(self.num_neurons)
        self.neg_beta = torch.rand(self.num_neurons)
        self.pos_gamma = torch.rand(self.num_neurons)
        self.neg_gamma = torch.rand(self.num_neurons)
        self.pos_threshold = torch.rand(self.num_neurons)
        self.neg_threshold = torch.rand(self.num_neurons)
        self.pos_scale = torch.randn(self.num_neurons)
        self.neg_scale = torch.randn(self.num_neurons)
        self.pos_rec_weight = torch.randn(self.num_neurons)
        self.neg_rec_weight = torch.randn(self.num_neurons)
        self.bias = torch.randn(self.num_neurons) / 3.0

    def test_li(self):
        self._test_layer_equivalence(
            lambda: snn.LI(
                self.num_neurons,
                beta=self.pos_beta,
            ),
            lambda: snn.flex.LI(
                self.num_neurons,
                beta=self.pos_beta,
            ),
            "LI"
        )

    def test_dli(self):
        self._test_layer_equivalence(
            lambda: snn.DLI(
                self.num_neurons,
                pos_beta=self.pos_beta,
                neg_beta=self.neg_beta,
            ),
            lambda: snn.flex.DLI(
                self.num_neurons,
                pos_beta=self.pos_beta,
                neg_beta=self.neg_beta,
            ),
            "DLI"
        )

    def test_sli(self):
        self._test_layer_equivalence(
            lambda: snn.SLI(
                self.num_neurons,
                alpha=self.pos_alpha,
                beta=self.pos_beta,
            ),
            lambda: snn.flex.SLI(
                self.num_neurons,
                alpha=self.pos_alpha,
                beta=self.pos_beta,
            ),
            "SLI"
        )

    def test_lib(self):
        self._test_layer_equivalence(
            lambda: snn.LIB(
                self.num_neurons,
                beta=self.pos_beta,
                threshold=self.pos_threshold,
            ),
            lambda: snn.flex.LIB(
                self.num_neurons,
                beta=self.pos_beta,
                threshold=self.pos_threshold,
            ),
            layer_name="LIB"
        )

    def _test_layer_equivalence(self, base_layer_factory, flex_layer_factory, layer_name):
        """Test equivalence with simpler approach"""

        # Create layers
        base_layer = base_layer_factory()
        flex_layer = flex_layer_factory()

        # Create optimizers
        base_opt = torch.optim.Adam(base_layer.parameters(), lr=self.lr)
        flex_opt = torch.optim.Adam(flex_layer.parameters(), lr=self.lr)

        # Test 1: Forward pass equivalence
        base_layer.zero_states()
        flex_layer.zero_states()

        with torch.no_grad():
            for t in range(self.num_timesteps):
                # Create fresh input for each timestep to avoid graph issues
                x = torch.randn(self.batch_size, self.num_neurons)

                base_out = base_layer(x)
                flex_out = flex_layer(x)

                torch.testing.assert_close(
                    base_out, flex_out,
                    rtol=1e-5, atol=1e-6,
                    msg=f"{layer_name} forward pass mismatch at timestep {t}\ndiff:{base_out - flex_out}"
                )

        # Test 2: Training equivalence (simpler approach)
        base_layer.zero_states()
        flex_layer.zero_states()

        for t in range(self.num_timesteps):
            # Fresh input and target
            x = torch.randn(self.batch_size, self.num_neurons)
            target = torch.randn(self.batch_size, self.num_neurons)

            # Forward pass
            base_out = base_layer(x)
            flex_out = flex_layer(x)

            # Check forward equivalence during training
            torch.testing.assert_close(
                base_out, flex_out,
                rtol=1e-5, atol=1e-6,
                msg=f"{layer_name} training forward mismatch at step {t}\ndiff:{base_out - flex_out}"
            )

            # Calculate loss
            base_loss = nn.functional.mse_loss(base_out, target)
            flex_loss = nn.functional.mse_loss(flex_out, target)

            # Backward pass
            base_layer.zero_grad()
            flex_layer.zero_grad()
            base_loss.backward()
            flex_loss.backward()

            base_layer.detach_states()
            flex_layer.detach_states()

            # Update parameters
            base_opt.step()
            flex_opt.step()



from typing import TypedDict, Optional, Literal, Union, Dict, Any
from ._leaky_integrator import LeakyIntegrator
from ... import functional
import torch
from torch import nn


class LIB(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            beta: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            beta_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            learn_beta: bool = True,
            learn_threshold: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        beta_setup = {
            "value": beta,
            "rank": beta_rank,
            "ema": True,
            "learnable": learn_beta,
        }

        threshold_setup = {
            "value": threshold,
            "rank": threshold_rank,
            "surrogate": surrogate_derivative,
            "learnable": learn_threshold,
        }

        super().__init__(
            num_neurons=num_neurons,
            dim=dim,
            pos_beta_setup=beta_setup,
            pos_threshold_setup=threshold_setup,
        )


class DLIB(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
            learn_threshold: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass


class SLIB(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            alpha: Union[float, torch.Tensor] = 0.5,
            beta: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            alpha_rank: Literal[0, 1] = 1,
            beta_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            learn_alpha: bool = True,
            learn_beta: bool = True,
            learn_threshold: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass


class RLIB(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            beta: Union[float, torch.Tensor] = 0.9,
            gamma: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            rec_weight: Union[float, torch.Tensor] = 0.0,
            dim: int = -1,
            beta_rank: Literal[0, 1] = 1,
            gamma_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            rec_weight_rank: Literal[0, 1] = 1,
            learn_beta: bool = True,
            learn_gamma: bool = True,
            learn_threshold: bool = True,
            learn_rec_weight: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass


class DSLIB(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            pos_alpha: Union[float, torch.Tensor] = 0.5,
            neg_alpha: Union[float, torch.Tensor] = 0.5,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            pos_alpha_rank: Literal[0, 1] = 1,
            neg_alpha_rank: Literal[0, 1] = 1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            learn_pos_alpha: bool = True,
            learn_neg_alpha: bool = True,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
            learn_threshold: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass


class DRLIB(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            pos_gamma: Union[float, torch.Tensor] = 0.9,
            neg_gamma: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            pos_rec_weight: Union[float, torch.Tensor] = 0.0,
            neg_rec_weight: Union[float, torch.Tensor] = 0.0,
            dim: int = -1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            pos_gamma_rank: Literal[0, 1] = 1,
            neg_gamma_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            pos_rec_weight_rank: Literal[0, 1] = 1,
            neg_rec_weight_rank: Literal[0, 1] = 1,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
            learn_pos_gamma: bool = True,
            learn_neg_gamma: bool = True,
            learn_threshold: bool = True,
            learn_pos_rec_weight: bool = True,
            learn_neg_rec_weight: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass


class SRLIB(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            alpha: Union[float, torch.Tensor] = 0.5,
            beta: Union[float, torch.Tensor] = 0.9,
            gamma: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            rec_weight: Union[float, torch.Tensor] = 0.0,
            dim: int = -1,
            alpha_rank: Literal[0, 1] = 1,
            beta_rank: Literal[0, 1] = 1,
            gamma_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            rec_weight_rank: Literal[0, 1] = 1,
            learn_alpha: bool = True,
            learn_beta: bool = True,
            learn_gamma: bool = True,
            learn_threshold: bool = True,
            learn_rec_weight: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass


class DSRLIB(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            pos_alpha: Union[float, torch.Tensor] = 0.5,
            neg_alpha: Union[float, torch.Tensor] = 0.5,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            pos_gamma: Union[float, torch.Tensor] = 0.9,
            neg_gamma: Union[float, torch.Tensor] = 0.9,
            threshold: Union[float, torch.Tensor] = 1.0,
            pos_rec_weight: Union[float, torch.Tensor] = 0.0,
            neg_rec_weight: Union[float, torch.Tensor] = 0.0,
            dim: int = -1,
            pos_alpha_rank: Literal[0, 1] = 1,
            neg_alpha_rank: Literal[0, 1] = 1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            pos_gamma_rank: Literal[0, 1] = 1,
            neg_gamma_rank: Literal[0, 1] = 1,
            threshold_rank: Literal[0, 1] = 1,
            pos_rec_weight_rank: Literal[0, 1] = 1,
            neg_rec_weight_rank: Literal[0, 1] = 1,
            learn_pos_alpha: bool = True,
            learn_neg_alpha: bool = True,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
            learn_pos_gamma: bool = True,
            learn_neg_gamma: bool = True,
            learn_threshold: bool = True,
            learn_pos_rec_weight: bool = True,
            learn_neg_rec_weight: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass



import torch
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.colors import Normalize
from pathlib import Path
import math


def render_image(tensor: torch.Tensor, title: str = None, name: str = "", save: bool = False):
    B, C, H, W = tensor.shape

    cols = math.ceil(math.sqrt(B))
    rows = math.ceil(B / cols)

    fig, axes = plt.subplots(rows, cols, figsize=(cols * 2, rows * 2))
    axes = axes.flatten() if B > 1 else [axes]

    for i in range(rows * cols):
        ax = axes[i]
        if i < B:
            img = tensor[i]
            if C == 1:
                img = img.squeeze(0)
                ax.imshow(img.cpu(), cmap='gray')
            elif C == 3:
                img = img.permute(1, 2, 0)
                ax.imshow(img.cpu())
            else:
                raise ValueError(f"Unsupported number of channels: {C}")
        ax.axis('off')

    if title:
        fig.suptitle(title, fontsize=16)
        plt.subplots_adjust(top=0.9)

    plt.tight_layout()
    if save:
        Path("media").mkdir(parents=True, exist_ok=True)
        plt.savefig(f"media/{name}.png")
    plt.show()



from typing import TypedDict, Optional, Literal, Union, Dict, Any
import torch
from torch import nn
from ._tt_infrastructure import TTLayer


class LI(TTLayer):
    def __init__(
            self,
            num_neurons: int,
            beta: Union[float, torch.Tensor] = 0.9,
            dim: int = -1,
            beta_rank: Literal[0, 1] = 1,
            learn_beta: bool = True,
    ):
        super().__init__(num_neurons, dim)

        self._initialize_state("mem")
        self._register_decay("beta", beta, beta_rank, learn_beta)

    def forward(self, x):
        self._ensure_states(x)
        x = self._to_working_dim(x)

        mem = self._to_working_dim(self.mem)
        mem = mem * self.beta + x

        self.mem = self._from_working_dim(mem)

        return self.mem


class DLI(TTLayer):
    def __init__(
            self,
            num_neurons: int,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            dim: int = -1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
    ):
        super().__init__(num_neurons, dim)

        self._initialize_state("pos_mem")
        self._initialize_state("neg_mem")
        self._register_decay("pos_beta", pos_beta, pos_beta_rank, learn_pos_beta)
        self._register_decay("neg_beta", neg_beta, neg_beta_rank, learn_neg_beta)

    def forward(self, x):
        self._ensure_states(x)
        x = self._to_working_dim(x)

        pos_mem = self._to_working_dim(self.pos_mem)
        neg_mem = self._to_working_dim(self.neg_mem)
        pos_mem = pos_mem * self.pos_beta + torch.where(x >= 0, x, 0.0)
        neg_mem = neg_mem * self.neg_beta + torch.where(x <= 0, x, 0.0)

        self.pos_mem = self._from_working_dim(pos_mem)
        self.neg_mem = self._from_working_dim(neg_mem)

        mem = self.pos_mem + self.neg_mem

        return mem


class SLI(TTLayer):
    def __init__(
            self,
            num_neurons: int,
            alpha: Union[float, torch.Tensor] = 0.5,
            beta: Union[float, torch.Tensor] = 0.9,
            dim: int = -1,
            alpha_rank: Literal[0, 1] = 1,
            beta_rank: Literal[0, 1] = 1,
            learn_alpha: bool = True,
            learn_beta: bool = True,
    ):
        super().__init__(num_neurons, dim)

        self._initialize_state("syn")
        self._register_decay("alpha", alpha, alpha_rank, learn_alpha)

        self._initialize_state("mem")
        self._register_decay("beta", beta, beta_rank, learn_beta)

    def forward(self, x):
        self._ensure_states(x)
        x = self._to_working_dim(x)

        syn = self._to_working_dim(self.syn)
        syn = syn * self.alpha + x * (1 - self.alpha)

        mem = self._to_working_dim(self.mem)
        mem = mem * self.beta + syn

        self.syn = self._from_working_dim(syn)
        self.mem = self._from_working_dim(mem)

        return self.mem


class DSLI(TTLayer):
    def __init__(
            self,
            num_neurons: int,
            pos_alpha: Union[float, torch.Tensor] = 0.5,
            neg_alpha: Union[float, torch.Tensor] = 0.5,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            dim: int = -1,
            pos_alpha_rank: Literal[0, 1] = 1,
            neg_alpha_rank: Literal[0, 1] = 1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            learn_pos_alpha: bool = True,
            learn_neg_alpha: bool = True,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
    ):
        super().__init__(num_neurons, dim)

        self._initialize_state("pos_syn")
        self._initialize_state("neg_syn")
        self._register_decay("pos_alpha", pos_alpha, pos_alpha_rank, learn_pos_alpha)
        self._register_decay("neg_alpha", neg_alpha, neg_alpha_rank, learn_neg_alpha)

        self._initialize_state("pos_mem")
        self._initialize_state("neg_mem")
        self._register_decay("pos_beta", pos_beta, pos_beta_rank, learn_pos_beta)
        self._register_decay("neg_beta", neg_beta, neg_beta_rank, learn_neg_beta)

    def forward(self, x):
        self._ensure_states(x)

        x = self._to_working_dim(x)

        pos_syn = self._to_working_dim(self.pos_syn)
        neg_syn = self._to_working_dim(self.neg_syn)
        pos_syn = pos_syn * self.pos_alpha + torch.where(x >= 0, x, 0.0) * (1 - self.pos_alpha)
        neg_syn = neg_syn * self.neg_alpha + torch.where(x <= 0, x, 0.0) * (1 - self.neg_alpha)

        self.pos_syn = self._from_working_dim(pos_syn)
        self.neg_syn = self._from_working_dim(neg_syn)

        syn = pos_syn + neg_syn

        pos_mem = self._to_working_dim(self.pos_mem)
        neg_mem = self._to_working_dim(self.neg_mem)
        pos_mem = pos_mem * self.pos_beta + torch.where(syn >= 0, syn, 0.0)
        neg_mem = neg_mem * self.neg_beta + torch.where(syn <= 0, syn, 0.0)

        self.pos_mem = self._from_working_dim(pos_mem)
        self.neg_mem = self._from_working_dim(neg_mem)

        mem = self.pos_mem + self.neg_mem

        return mem



from typing import List, Dict, Any, Optional, Tuple
import torch
import numpy as np
import matplotlib.pyplot as plt


def _safe_to_numpy(t: torch.Tensor) -> np.ndarray:
	t = t.detach().cpu().ravel()
	a = t.numpy()
	# remove nans and infs
	a = a[np.isfinite(a)]
	return a


def _silverman_bandwidth(x: np.ndarray) -> float:
	n = x.size
	if n < 2:
		return 1.0
	s = np.std(x, ddof=1)
	# Silverman's rule of thumb for Gaussian kernel
	bw = 1.06 * s * n ** (-1 / 5)
	# guard against zero bandwidth
	if bw <= 0 or not np.isfinite(bw):
		bw = np.ptp(x) / 10.0 if np.ptp(x) > 0 else 1.0
	return bw


def _gaussian_kde_eval(samples: np.ndarray,
					   grid: np.ndarray,
					   bw: Optional[float] = None,
					   max_samples: int = 20000) -> np.ndarray:
	"""
	Vectorized Gaussian KDE evaluation at points in `grid`.
	- samples: 1D numpy array of data
	- grid: 1D numpy array of evaluation points
	- bw: bandwidth (if None uses Silverman)
	- max_samples: if samples too large, random subsample for speed
	Returns density values aligned with grid (integrates approximately to 1).
	"""
	if samples.size == 0:
		return np.zeros_like(grid, dtype=float)
	x = samples
	n = x.size
	if n > max_samples:
		rng = np.random.default_rng(0)
		x = rng.choice(x, size=max_samples, replace=False)
		n = x.size
	if bw is None:
		bw = _silverman_bandwidth(x)
	# compute kernel densities: sum of Gaussians
	# density = (1/(n*bw)) * sum_j phi((grid - x_j)/bw)
	# vectorized with broadcasting (grid[:,None] - x[None,:])
	diffs = (grid[:, None] - x[None, :]) / bw
	# gaussian kernel
	K = np.exp(-0.5 * diffs * diffs) / np.sqrt(2 * np.pi)
	dens = K.sum(axis=1) / (n * bw)
	# normalize numerical integration to 1 (improves comparability across grids)
	area = np.trapz(dens, grid)
	if area > 0:
		dens = dens / area
	return dens


def distributions(layers: List[torch.Tensor],
				  title: str,
				  *,
				  n_grid: int = 1024,
				  n_percentiles: int = 100,
				  bandwidths: Optional[List[Optional[float]]] = None,
				  show_percentile_lines: bool = False,
				  compute_metrics: bool = True,
				  max_kde_samples: int = 20000
				  ) -> Dict[str, Any]:
	"""
	Estimate and plot continuous distributions for each layer (torch.Tensor) in `layers`.
	Parameters
	----------
	title : str
		Title for the single overlay plot.
	layers : list[torch.Tensor]
		Each item is a tensor containing continuous values for that layer (will be flattened).
	n_grid : int
		Number of x points for KDE evaluation.
	n_percentiles : int
		Number of percentile buckets (e.g. 100 -> percentiles at 0,1,...,100).
	bandwidths : list or None
		Optional list of bandwidths per layer; None to use Silverman's rule.
	show_percentile_lines : bool
		If True, plots vertical lines for a few percentiles (can clutter plot).
	compute_metrics : bool
		If True attempts to compute pairwise similarity metrics and returns them.
	max_kde_samples : int
		Maximum number of samples per layer used to compute KDE (subsample if larger).
	Returns
	-------
	dict with keys:
	  - 'grid': evaluation grid (numpy array)
	  - 'kdes': list of density arrays (same order as layers)
	  - 'percentiles': list of percentile arrays per layer
	  - 'stats': list of dicts {n, mean, std, min, max}
	  - 'metrics': dict of pairwise metrics (if compute_metrics)
	:param title:
	:param layers:
	:param n_grid:
	:param n_percentiles:
	:param bandwidths:
	:param show_percentile_lines:
	:param compute_metrics:
	:param max_kde_samples:
	:return:
	"""
	# prepare data arrays
	arrs = [_safe_to_numpy(t) for t in layers]
	stats = []
	# compute global x-range for consistent plotting
	finite_arrs = [a for a in arrs if a.size > 0]
	if len(finite_arrs) == 0:
		raise ValueError("No finite values found in any input tensor.")
	global_min = min(a.min() for a in finite_arrs)
	global_max = max(a.max() for a in finite_arrs)
	# expand a bit for display
	pad = 0.01 * (global_max - global_min) if global_max > global_min else 1.0
	x_min = global_min - pad
	x_max = global_max + pad
	grid = np.linspace(x_min, x_max, n_grid, dtype=float)

	if bandwidths is None:
		bandwidths = [None] * len(arrs)
	elif len(bandwidths) != len(arrs):
		raise ValueError("If provided, bandwidths must have same length as layers")

	kdes = []
	percentiles = []
	for i, (a, bw) in enumerate(zip(arrs, bandwidths)):
		if a.size == 0:
			kdes.append(np.zeros_like(grid))
			percentiles.append(np.full(n_percentiles + 1, np.nan))
			stats.append({'n': 0, 'mean': np.nan, 'std': np.nan, 'min': np.nan, 'max': np.nan})
			continue
		dens = _gaussian_kde_eval(a, grid, bw=bw, max_samples=max_kde_samples)
		kdes.append(dens)
		pct_values = np.percentile(a, np.linspace(0, 100, n_percentiles + 1))
		percentiles.append(pct_values)
		stats.append({'n': int(a.size), 'mean': float(a.mean()), 'std': float(a.std(ddof=1)),
					  'min': float(a.min()), 'max': float(a.max())})

	# plotting
	fig, ax = plt.subplots(figsize=(10, 6))
	ax.set_title(title)
	ax.set_xlabel("value")
	ax.set_ylabel("density (KDE)")
	lines = []
	labels = []
	for i, dens in enumerate(kdes):
		# label contains index and base stats
		s = stats[i]
		label = f"layer {i} (n={s['n']}, μ={s['mean']:.3g}, σ={s['std']:.3g})"
		l, = ax.plot(grid, dens, label=label)
		lines.append(l)
		labels.append(label)
		if show_percentile_lines and np.isfinite(percentiles[i]).all():
			# show a few percentiles lightly (10th, 50th, 90th)
			for p in (10, 50, 90):
				v = np.percentile(arrs[i], p)
				ax.axvline(v, linestyle='--', linewidth=0.6, alpha=0.6)

	ax.legend(loc='best', fontsize='small')
	ax.grid(True, linestyle=':', alpha=0.4)
	plt.tight_layout()
	plt.show()

	result: Dict[str, Any] = {
		'grid': grid,
		'kdes': kdes,
		'percentiles': percentiles,
		'stats': stats,
		'fig_ax': (fig, ax)  # note: matplotlib objects included for programmatic use
	}

	# compute pairwise similarity metrics
	if compute_metrics:
		# try to use scipy for Wasserstein and KS if available; otherwise fall back to grid L1
		pairwise_wasserstein = None
		pairwise_l1 = np.zeros((len(arrs), len(arrs)), dtype=float)
		try:
			from scipy.stats import wasserstein_distance
			use_wasserstein = True
			pairwise_wasserstein = np.zeros((len(arrs), len(arrs)), dtype=float)
		except Exception:
			use_wasserstein = False
			pairwise_wasserstein = None

		# L1 between normalized KDEs on the grid: integral |f - g|
		for i in range(len(arrs)):
			for j in range(len(arrs)):
				f = kdes[i]
				g = kdes[j]
				pairwise_l1[i, j] = np.trapz(np.abs(f - g), grid)
				if use_wasserstein:
					# if one of the arrays empty, put nan
					a = arrs[i]
					b = arrs[j]
					if a.size == 0 or b.size == 0:
						pairwise_wasserstein[i, j] = np.nan
					else:
						pairwise_wasserstein[i, j] = wasserstein_distance(a, b)

		metrics = {'l1_kde': pairwise_l1}
		if use_wasserstein:
			metrics['wasserstein'] = pairwise_wasserstein
		result['metrics'] = metrics

	return result



import torch


def sigmoid_inverse(x: torch.Tensor) -> torch.Tensor:
	assert torch.all((x > 0) & (x < 1)), "Tensor for inverse sigmoid has values outside the (0,1) range"
	return torch.logit(x)


def softplus_inverse(x: torch.Tensor) -> torch.Tensor:
	assert torch.all(x > 0), "Tensor for softplus inverse has values outside the (0,inf) range"
	return torch.log(torch.expm1(x))



from typing import TypedDict, Optional, Literal, Union, Dict, Any
import torch
from torch import nn
from .._tt_infrastructure import TTLayer
from ...functional import atan_surrogate, sigmoid_surrogate


class DecayConfig(TypedDict, total=False):
    value: Union[float, torch.Tensor]
    rank: Literal[0, 1]
    ema: bool
    learnable: bool


class ThresholdConfig(TypedDict, total=False):
    value: Union[float, torch.Tensor]
    rank: Literal[0, 1]
    surrogate: Any  # has to be a surrogate function
    passthrough: bool
    learnable: bool


class VectorConfig(TypedDict, total=False):
    value: Union[float, torch.Tensor]
    rank: Literal[0, 1]
    learnable: bool


DEFAULT_ALPHA = {"value": 0.5, "rank": 1, "ema": True, "learnable": True}
DEFAULT_BETA = {"value": 0.9, "rank": 1, "ema": False, "learnable": True}
DEFAULT_GAMMA = {"value": 0.9, "rank": 1, "ema": True, "learnable": True}
DEFAULT_THRESHOLD = {"value": 1.0, "rank": 1, "surrogate": atan_surrogate(2.0), "passthrough": False, "learnable": True}
DEFAULT_SCALE = {"value": 1.0, "rank": 1, "learnable": True}
DEFAULT_REC_WEIGHT = {"value": 0.0, "rank": 1, "learnable": True}
DEFAULT_BIAS = {"value": 0.0, "rank": 1, "learnable": True}


class LeakyIntegrator(TTLayer):
    def __init__(
            self,
            num_neurons: int,
            dim: int = -1,

            pos_alpha_setup: Optional[DecayConfig] = None,
            neg_alpha_setup: Optional[DecayConfig] = None,
            pos_beta_setup: Optional[DecayConfig] = DEFAULT_BETA,
            neg_beta_setup: Optional[DecayConfig] = None,
            pos_gamma_setup: Optional[DecayConfig] = None,
            neg_gamma_setup: Optional[DecayConfig] = None,
            pos_threshold_setup: Optional[ThresholdConfig] = None,
            neg_threshold_setup: Optional[ThresholdConfig] = None,
            pos_scale_setup: Optional[VectorConfig] = None,
            neg_scale_setup: Optional[VectorConfig] = None,
            pos_rec_weight_setup: Optional[VectorConfig] = None,
            neg_rec_weight_setup: Optional[VectorConfig] = None,
            bias_setup: Optional[VectorConfig] = None,

    ):
        super().__init__(num_neurons, dim)

        def _setup_decay(
                name: str,
                pos_setup: Optional[DecayConfig],
                neg_setup: Optional[DecayConfig],
                default_config: dict
        ):
            setattr(self, f"use_{name}", pos_setup is not None or neg_setup is not None)

            if getattr(self, f"use_{name}"):
                setattr(self, f"dual_{name}", pos_setup is not None and neg_setup is not None)
                state_mapping = {"alpha": "syn", "beta": "mem", "gamma": "rec"}
                state_name = state_mapping[name]
                if getattr(self, f"dual_{name}"):  # if duality
                    self._initialize_state(f"pos_{state_name}")
                    self._initialize_state(f"neg_{state_name}")
                    pos_cfg = {**default_config, **(pos_setup or {})}
                    neg_cfg = {**default_config, **(neg_setup or {})}
                    self._register_decay(f"pos_{name}", pos_cfg["value"], pos_cfg["rank"], pos_cfg["learnable"])
                    self._register_decay(f"neg_{name}", neg_cfg["value"], neg_cfg["rank"], neg_cfg["learnable"])
                    setattr(self, f"ema_pos_{state_name}", pos_cfg["ema"])
                    setattr(self, f"ema_neg_{state_name}", pos_cfg["ema"])
                else:
                    self._initialize_state(f"{state_name}")
                    setup = pos_setup or neg_setup
                    cfg = {**default_config, **(setup or {})}
                    self._register_decay(f"{name}", cfg["value"], cfg["rank"], cfg["learnable"])
                    setattr(self, f"ema_{state_name}", cfg["ema"])

        _setup_decay("alpha", pos_alpha_setup, neg_alpha_setup, DEFAULT_ALPHA)
        _setup_decay("beta", pos_beta_setup, neg_beta_setup, DEFAULT_BETA)
        _setup_decay("gamma", pos_gamma_setup, neg_gamma_setup, DEFAULT_GAMMA)

        assert self.use_beta, "beta must be initialized"

        if self.use_gamma:
            self._initialize_state("prev_output")

        def _setup_threshold(
                name: str,
                setup: Optional[DecayConfig],
                default_config: dict
        ):
            cfg = {**default_config, **(setup or {})}
            setattr(self, f"use_{name}", setup is not None)
            if getattr(self, f"use_{name}"):
                setattr(self, f"passthrough_{name}", cfg["passthrough"])
                if not getattr(self, f"passthrough_{name}"):  # if not passthrough
                    self._register_threshold(f"{name}", cfg["value"], cfg["rank"], cfg["learnable"])
                    setattr(self, f"{name}_heaviside", cfg["surrogate"])

        _setup_threshold("pos_threshold", pos_threshold_setup, DEFAULT_THRESHOLD)
        _setup_threshold("neg_threshold", neg_threshold_setup, DEFAULT_THRESHOLD)

        def _setup_vector(
                name: str,
                pos_setup: Optional[DecayConfig],
                neg_setup: Optional[DecayConfig],
                default_config: dict
        ):
            setattr(self, f"use_{name}", pos_setup is not None or neg_setup is not None)
            if getattr(self, f"use_{name}"):
                setattr(self, f"dual_{name}", pos_setup is not None and neg_setup is not None)
                if getattr(self, f"dual_{name}"):
                    pos_cfg = {**default_config, **(pos_setup or {})}
                    neg_cfg = {**default_config, **(neg_setup or {})}
                    self._register_parameter(f"pos_{name}", pos_cfg["value"], pos_cfg["rank"], pos_cfg["learnable"])
                    self._register_parameter(f"neg_{name}", neg_cfg["value"], neg_cfg["rank"], neg_cfg["learnable"])
                else:
                    setup = pos_setup or neg_setup
                    cfg = {**default_config, **(setup or {})}
                    self._register_parameter(f"{name}", cfg["value"], cfg["rank"], cfg["learnable"])

        _setup_vector("scale", pos_scale_setup, neg_scale_setup, DEFAULT_SCALE)
        _setup_vector("rec_weight", pos_rec_weight_setup, neg_rec_weight_setup, DEFAULT_REC_WEIGHT)
        _setup_vector("bias", bias_setup, None, DEFAULT_BIAS)

        if self.use_rec_weight:
            assert self.use_gamma, "rec_weight is initialized, but gamma decay and rec trace is not"

    def forward(self, x):
        self._ensure_states(x)
        x_moved = self._to_working_dim(x)

        if self.use_alpha:
            if self.dual_alpha:
                pos_syn_moved = self._to_working_dim(self.pos_syn)
                neg_syn_moved = self._to_working_dim(self.neg_syn)

                pos_syn_moved_delta = torch.where(x_moved >= 0, x_moved, 0.0)
                neg_syn_moved_delta = torch.where(x_moved <= 0, x_moved, 0.0)

                if self.ema_pos_syn:
                    pos_syn_moved_delta = pos_syn_moved_delta * (1 - self.pos_alpha)

                if self.ema_neg_syn:
                    neg_syn_moved_delta = neg_syn_moved_delta * (1 - self.neg_alpha)

                pos_syn_moved = pos_syn_moved * self.pos_alpha + pos_syn_moved_delta
                neg_syn_moved = neg_syn_moved * self.neg_alpha + neg_syn_moved_delta

                mem_delta = pos_syn_moved + neg_syn_moved

                self.pos_syn = self._from_working_dim(pos_syn_moved)
                self.neg_syn = self._from_working_dim(neg_syn_moved)
            else:
                syn_moved = self._to_working_dim(self.syn)

                syn_moved_delta = x_moved

                if self.ema_syn:
                    syn_moved_delta = syn_moved_delta * (1 - self.alpha)

                syn_moved = syn_moved * self.alpha + syn_moved_delta

                mem_delta = syn_moved

                self.syn = self._from_working_dim(syn_moved)
        else:
            mem_delta = x_moved

        if self.use_gamma:
            prev_output_moved = self._to_working_dim(self.prev_output)
            if self.dual_gamma:
                pos_rec_moved = self._to_working_dim(self.pos_rec)
                neg_rec_moved = self._to_working_dim(self.neg_rec)

                pos_rec_moved_delta = torch.where(prev_output_moved >= 0, prev_output_moved, 0.0)
                neg_rec_moved_delta = torch.where(prev_output_moved <= 0, prev_output_moved, 0.0)

                # rec_weight scaling must happen BEFORE integration into mem_delta, for mathematical reasons
                if self.use_rec_weight:
                    if self.dual_rec_weight:
                        pos_rec_moved_delta = pos_rec_moved_delta * self.pos_rec_weight
                        neg_rec_moved_delta = neg_rec_moved_delta * self.neg_rec_weight
                    else:
                        pos_rec_moved_delta = pos_rec_moved_delta * self.rec_weight
                        neg_rec_moved_delta = neg_rec_moved_delta * self.rec_weight

                if self.ema_pos_rec:
                    pos_rec_moved_delta = pos_rec_moved_delta * (1 - self.pos_gamma)

                if self.ema_neg_rec:
                    neg_rec_moved_delta = neg_rec_moved_delta * (1 - self.neg_gamma)

                pos_rec_moved = pos_rec_moved * self.pos_gamma + pos_rec_moved_delta
                neg_rec_moved = neg_rec_moved * self.neg_gamma + neg_rec_moved_delta

                mem_delta = mem_delta + pos_rec_moved + neg_rec_moved

                self.pos_rec = self._from_working_dim(pos_rec_moved)
                self.neg_rec = self._from_working_dim(neg_rec_moved)
            else:
                rec_moved = self._to_working_dim(self.rec)

                # rec_weight scaling must happen BEFORE integration into mem_delta, for mathematical reasons
                if self.use_rec_weight:
                    if self.dual_rec_weight:
                        pos_rec_moved_delta = torch.where(prev_output_moved >= 0, prev_output_moved, 0.0)
                        neg_rec_moved_delta = torch.where(prev_output_moved <= 0, prev_output_moved, 0.0)

                        pos_rec_moved_delta = pos_rec_moved_delta * self.pos_rec_weight
                        neg_rec_moved_delta = neg_rec_moved_delta * self.neg_rec_weight

                        rec_moved_delta = pos_rec_moved_delta + neg_rec_moved_delta
                    else:
                        rec_moved_delta = prev_output_moved * self.rec_weight

                if self.ema_rec:
                    rec_moved_delta = rec_moved_delta * (1 - self.gamma)

                rec_moved = rec_moved * self.gamma + rec_moved_delta

                mem_delta = mem_delta + rec_moved

                self.rec = self._from_working_dim(rec_moved)

        if self.use_bias:
            mem_delta = mem_delta + self.bias

        # ASSUMED THAT BETA IS USED SINCE ASSERTED AT INIT
        if self.dual_beta:
            pos_mem_moved = self._to_working_dim(self.pos_mem)
            neg_mem_moved = self._to_working_dim(self.neg_mem)

            pos_mem_moved_delta = torch.where(mem_delta >= 0, mem_delta, 0.0)
            neg_mem_moved_delta = torch.where(mem_delta <= 0, mem_delta, 0.0)

            if self.ema_pos_mem:
                pos_mem_moved_delta = pos_mem_moved_delta * (1 - self.pos_beta)

            if self.ema_neg_mem:
                neg_mem_moved_delta = neg_mem_moved_delta * (1 - self.neg_beta)

            pos_mem_moved = pos_mem_moved * self.pos_beta + pos_mem_moved_delta
            neg_mem_moved = neg_mem_moved * self.neg_beta + neg_mem_moved_delta

            mem_moved = pos_mem_moved + neg_mem_moved
        else:
            mem_moved = self._to_working_dim(self.mem)

            if self.ema_mem:
                mem_delta = mem_delta * (1 - self.beta)

            mem_moved = mem_moved * self.beta + mem_delta

        if self.use_pos_threshold or self.use_neg_threshold:
            output = torch.zeros_like(x)
            output_moved = self._to_working_dim(output)
            if self.use_pos_threshold:
                if self.passthrough_pos_threshold:
                    pos_spikes = torch.where(mem_moved >= 0, mem_moved, 0.0)
                else:
                    pos_spikes = self.pos_threshold_heaviside(mem_moved - self.pos_threshold)
                output_moved = output_moved + pos_spikes

            if self.use_neg_threshold:
                if self.passthrough_neg_threshold:
                    neg_spikes = torch.where(mem_moved <= 0, mem_moved, 0.0)
                else:
                    neg_spikes = -self.neg_threshold_heaviside(self.neg_threshold - mem_moved)
                output_moved = output_moved + neg_spikes

            if self.use_pos_threshold and not self.passthrough_pos_threshold:
                threshold_subtraction = pos_spikes * self.pos_threshold
                if self.dual_beta:
                    pos_mem_moved = pos_mem_moved - threshold_subtraction * 0.5
                    neg_mem_moved = neg_mem_moved - threshold_subtraction * 0.5
                else:
                    mem_moved = mem_moved - threshold_subtraction

            if self.use_neg_threshold and not self.passthrough_neg_threshold:
                threshold_subtraction = neg_spikes * self.neg_threshold
                if self.dual_beta:
                    pos_mem_moved = pos_mem_moved - threshold_subtraction * 0.5
                    neg_mem_moved = neg_mem_moved - threshold_subtraction * 0.5
                else:
                    mem_moved = mem_moved - threshold_subtraction
        else:
            output_moved = mem_moved

        # to decouple scale from rec_weight, these things happen separately
        if self.use_gamma:
            self.prev_output = self._from_working_dim(output_moved)

        if self.use_scale:
            if self.dual_scale:
                output_moved = output_moved * torch.where(output_moved > 0, self.pos_scale, 1.0)
                output_moved = output_moved * torch.where(output_moved < 0, self.neg_scale, 1.0)
            else:
                output_moved = output_moved * self.scale

        output = self._from_working_dim(output_moved)

        if self.dual_beta:
            self.pos_mem = self._from_working_dim(pos_mem_moved)
            self.neg_mem = self._from_working_dim(neg_mem_moved)
        else:
            self.mem = self._from_working_dim(mem_moved)

        return output



import math


def halflife_to_decay(num_steps: float):
    return math.exp(-math.log(2) / num_steps)


def decay_to_halflife(decay: float):
    return -math.log(2) / math.log(decay)


def timesteps_to_decay(time: float):
    return 1 - 1 / time


def decay_to_timesteps(decay: float):
    return 1 / (1 - decay)



import torch
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LinearSegmentedColormap


def create_diverging_colormap():
    """Create a colormap with blue for positive, red for negative values."""
    colors = ['red', 'white', 'blue']  # negative -> white -> positive
    n_bins = 256
    cmap = LinearSegmentedColormap.from_list('diverging', colors, N=n_bins)
    return cmap


def spike_train(
        list_of_tensors,
        spacing: float = 1.0,
        linelength: float = 0.8,
        linewidth: float = 0.5,
        title: str = "Spike Train Raster",
        use_imshow: bool = True,
):
    # Stack into 2D array: shape (num_neurons, T)
    # list_of_tensors assumed length T each
    data = torch.stack(list_of_tensors).cpu().detach().numpy()  # shape (T, N)
    data = data.T  # shape (N, T)

    plt.figure(figsize=(8, 4))
    if use_imshow:
        # Create custom colormap for diverging values
        cmap = create_diverging_colormap()

        # Determine symmetric color scale
        vmax = max(abs(data.max()), abs(data.min()))
        vmin = -vmax

        plt.imshow(
            data,
            aspect='auto',
            cmap=cmap,
            origin='lower',
            interpolation='nearest',
            vmin=vmin,
            vmax=vmax
        )
        plt.colorbar(label='Spike value')
    else:
        N, T = data.shape
        # for each neuron i, list of spike times where value != 0
        spike_times = [
            [t for t in range(T) if data[i, t] != 0]
            for i in range(N)
        ]
        # y offsets
        offsets = np.arange(N) * spacing
        plt.eventplot(
            spike_times,
            orientation='horizontal',
            lineoffsets=offsets,
            linelengths=linelength,
            linewidths=linewidth,
            colors='k'
        )
        plt.ylim(-spacing, offsets[-1] + spacing)

    plt.xlabel("Time step")
    plt.ylabel("Neuron index")
    plt.title(title)
    plt.tight_layout()
    plt.show()



from typing import TypedDict, Optional, Literal, Union, Dict, Any, Set
import torch
from torch import nn
from .. import functional


class TTLayer(nn.Module):
    """A mixin helper class, used to help manage parameters and hidden states."""

    def __init__(self, num_neurons: int, dim: int = -1):
        super().__init__()
        self._state_names = set()
        self.num_neurons = num_neurons
        self.dim = dim
        self.round_ste = functional.round_ste()
        self.bernoulli_ste = functional.bernoulli_ste()
        self.probabilistic_ste = functional.probabilistic_ste()

    def _register_parameter(
            self,
            name: str,
            value: Union[float, torch.Tensor],
            rank: Literal[0, 1],
            learnable: bool,
            inverse_function=lambda x: x,
            activation_function=lambda x: x,
    ):
        if isinstance(value, torch.Tensor):
            if value.ndim == 0:
                pass
            elif value.ndim == 1:
                assert value.numel() == self.num_neurons, f"{name} does not have {self.num_neurons} elements"
            else:
                raise ValueError(f"rank (.ndim) of provided {name} is not 0 (scalar) or 1 (vector)")
            param_tensor = inverse_function(value)
        else:
            value = float(value)
            if rank == 0:
                param_tensor = inverse_function(torch.tensor(value))
            elif rank == 1:
                param_tensor = inverse_function(torch.full([self.num_neurons], value))
            else:
                raise ValueError(f"{name} rank is not 0 (scalar) or 1 (vector)")

        # save the raw_ variant of the parameter as an nn.Parameter or buffer
        if learnable:
            setattr(self, f"raw_{name}", nn.Parameter(param_tensor.detach().clone()))
        else:
            self.register_buffer(f"raw_{name}", param_tensor.detach().clone())

        # create a @property of the raw_ parameter that passes it through the respective activation function
        setattr(self.__class__, name, self._make_property(name, activation_function))

    @staticmethod
    def _make_property(name, activation_function):
        def getter(self):
            return activation_function(getattr(self, f"raw_{name}"))

        return property(getter)

    def _register_decay(
            self,
            name: str,
            value: Union[float, torch.Tensor],
            rank: Literal[0, 1],
            learnable: bool,
    ):
        self._register_parameter(
            name,
            value,
            rank,
            learnable,
            inverse_function=functional.sigmoid_inverse,
            activation_function=nn.functional.sigmoid,
        )

    def _register_threshold(
            self,
            name: str,
            value: Union[float, torch.Tensor],
            rank: Literal[0, 1],
            learnable: bool,
    ):
        self._register_parameter(
            name,
            value,
            rank,
            learnable,
            inverse_function=functional.softplus_inverse,
            activation_function=nn.functional.softplus,
        )

    def _initialize_state(self, state_name: str):
        """Initialize and register a state name for bulk operations"""
        self._state_names.add(state_name)
        setattr(self, state_name, None)

    def _detach_state(self, state_name: str):
        """Detach a state tensor if it exists"""
        state = getattr(self, state_name)
        if state is not None:
            setattr(self, state_name, state.detach())

    def detach_states(self):
        """Detach all registered states"""
        for state_name in self._state_names:
            self._detach_state(state_name)

    def _zero_state(self, state_name: str):
        """Set a state to None"""
        setattr(self, state_name, None)

    def zero_states(self):
        """Zero all registered states"""
        for state_name in self._state_names:
            self._zero_state(state_name)

    def _ensure_state(self, state_name: str, reference_tensor: torch.Tensor):
        """Initialize a state with zeros_like if None, otherwise return existing state"""
        state = getattr(self, state_name)
        if state is None:
            state = torch.zeros_like(reference_tensor)
            setattr(self, state_name, state)

    def _ensure_states(self, reference_tensor: torch.Tensor):
        """Ensure all registered states are initialized"""
        for state_name in self._state_names:
            self._ensure_state(state_name, reference_tensor)

    def _to_working_dim(self, tensor: torch.Tensor) -> torch.Tensor:
        """Move tensor to working dimension (last dim)"""
        return tensor.movedim(self.dim, -1)

    def _from_working_dim(self, tensor: torch.Tensor) -> torch.Tensor:
        """Move tensor back from working dimension"""
        return tensor.movedim(-1, self.dim)

    def _count_params(self, total_list):
        """Append parameter count to the list."""
        total_list.append(sum(p.numel() for p in self.parameters()))

    def TTcompile(self):
        """Compile layer for inference by pre-computing parameters"""
        if hasattr(self, '_compiled') and self._compiled:
            return

        # Store compilation metadata for uncompile
        self._compile_metadata = {}

        # Find all raw_* parameters and their corresponding properties
        for attr_name in list(self.__dict__.keys()):
            if attr_name.startswith('raw_'):
                param_name = attr_name[4:]  # Remove 'raw_' prefix

                # Get current computed value
                computed_value = getattr(self, param_name)

                # Store metadata for uncompile
                self._compile_metadata[param_name] = {
                    'raw_value': getattr(self, attr_name).detach().clone(),
                    'is_parameter': isinstance(getattr(self, param_name), nn.Parameter),
                    'learnable': hasattr(self, f'learn_{param_name}') and getattr(self, f'learn_{param_name}')
                }

                # Replace raw parameter with pre-computed buffer
                delattr(self, attr_name)
                self.register_buffer(param_name, computed_value.detach().clone())

                # Remove the property (will be replaced by direct attribute access)
                if hasattr(self.__class__, param_name):
                    delattr(self.__class__, param_name)

        self._compiled = True

    def TTuncompile(self):
        """Uncompile layer to restore training capabilities"""
        if not hasattr(self, '_compiled') or not self._compiled:
            return

        # Restore raw_* parameters from metadata
        for param_name, metadata in self._compile_metadata.items():
            # Remove the compiled buffer
            if hasattr(self, param_name):
                delattr(self, param_name)

            # Restore raw parameter
            raw_name = f"raw_{param_name}"
            if metadata['learnable']:
                setattr(self, raw_name, nn.Parameter(metadata['raw_value']))
            else:
                self.register_buffer(raw_name, metadata['raw_value'])

            # Recreate the property (this is tricky - need to recreate the original property)
            # We'll need to store the activation function type during initial registration

        # Clean up
        delattr(self, '_compiled')
        delattr(self, '_compile_metadata')


class TTModel(nn.Module):
    """
    Base class that makes it trivial to call lifecycle methods (zero_states, detach_states, ...)
    across an entire model tree (registered submodules *and* common non-module containers).
    Inherit your models from this to get model.zero_states() / model.detach_states() behavior.
    """

    def __init__(self):
        super().__init__()

    def get_param_count(self) -> int:
        """Count learnable parameters only from SNN layers (TTLayers)."""
        param_counts = []
        self._call_recursive("_count_params", param_counts)
        return sum(param_counts)

    def save_states(self) -> Dict[str, torch.Tensor]:
        """Save all hidden states from TTLayers in the model.

        states = model.save_states()
        torch.save(states, "model_states.pt")

        Returns:
            Dictionary mapping layer_state_name -> tensor, compatible with torch.save()
        """
        states = {}

        def collect_states(obj, path=""):
            if isinstance(obj, TTLayer):
                for state_name in obj._state_names:
                    state_value = getattr(obj, state_name)
                    if state_value is not None:
                        # Use dot notation for unique identification
                        full_name = f"{path}.{state_name}" if path else state_name
                        states[full_name] = state_value.detach().clone()

            # Recurse into submodules and containers (similar to _call_recursive)
            if isinstance(obj, nn.Module):
                for name, child in obj._modules.items():
                    child_path = f"{path}.{name}" if path else name
                    collect_states(child, child_path)

            # Handle container attributes
            try:
                attrs = getattr(obj, "__dict__", {})
            except Exception:
                attrs = {}

            for attr_name, attr_value in attrs.items():
                if attr_value is None:
                    continue
                if isinstance(attr_value, (list, tuple, set)):
                    for i, el in enumerate(attr_value):
                        collect_states(el, f"{path}.{attr_name}[{i}]")
                elif isinstance(attr_value, dict):
                    for key, el in attr_value.items():
                        collect_states(el, f"{path}.{attr_name}[{key}]")
                elif isinstance(attr_value, (nn.Module, object)) and not isinstance(attr_value, TTLayer):
                    collect_states(attr_value, f"{path}.{attr_name}")

        collect_states(self)
        return states

    def load_states(self, states: Dict[str, torch.Tensor], strict: bool = True, device=None) -> None:
        """Load hidden states into TTLayers.

        states = torch.load("model_states.pt")
        model.load_states(states)

        Args:
            states: Dictionary from save_states() or torch.load()
            strict: If True, raise error for missing/extra states
            device: Target device for loaded states (auto-detected if None)
        """
        loaded_count = 0
        missing_states = []

        def distribute_states(obj, path=""):
            nonlocal loaded_count

            if isinstance(obj, TTLayer):
                for state_name in obj._state_names:
                    full_name = f"{path}.{state_name}" if path else state_name

                    if full_name in states:
                        state_tensor = states[full_name]
                        current_state = getattr(obj, state_name)

                        # Validate shape if current state exists
                        if current_state is not None:
                            if current_state.shape != state_tensor.shape:
                                raise ValueError(
                                    f"Shape mismatch for {full_name}: "
                                    f"expected {current_state.shape}, got {state_tensor.shape}"
                                )

                        # Use provided device, or detect from parameters/buffers, or fallback to current device
                        target_device = device
                        if target_device is None:
                            # Try to get device from existing parameters
                            for param in obj.parameters():
                                target_device = param.device
                                break
                            else:
                                # Try to get device from existing buffers
                                for buffer in obj.buffers():
                                    target_device = buffer.device
                                    break
                                else:
                                    # Fallback to the state tensor's device
                                    target_device = state_tensor.device

                        # Set the state, ensuring it's on the correct device
                        setattr(obj, state_name, state_tensor.to(target_device))
                        loaded_count += 1
                    else:
                        missing_states.append(full_name)

            # Recurse (same logic as save_states)
            if isinstance(obj, nn.Module):
                for name, child in obj._modules.items():
                    child_path = f"{path}.{name}" if path else name
                    distribute_states(child, child_path)

            try:
                attrs = getattr(obj, "__dict__", {})
            except Exception:
                attrs = {}

            for attr_name, attr_value in attrs.items():
                if attr_value is None:
                    continue
                if isinstance(attr_value, (list, tuple, set)):
                    for i, el in enumerate(attr_value):
                        distribute_states(el, f"{path}.{attr_name}[{i}]")
                elif isinstance(attr_value, dict):
                    for key, el in attr_value.items():
                        distribute_states(el, f"{path}.{attr_name}[{key}]")
                elif isinstance(attr_value, (nn.Module, object)) and not isinstance(attr_value, TTLayer):
                    distribute_states(attr_value, f"{path}.{attr_name}")

        distribute_states(self)

        if strict and missing_states:
            raise ValueError(f"Missing states for: {missing_states}")

        print(f"Loaded {loaded_count} states")

    def zero_states(self) -> None:
        """Public API — zero any stateful child that implements zero_states()."""
        self._call_recursive("zero_states")

    def detach_states(self) -> None:
        """Public API — detach (stop-gradient) any stateful child that implements detach_states()."""
        self._call_recursive("detach_states")

    # --- internal recursive walker ---
    def _call_recursive(self, method_name: str, *args, **kwargs) -> None:
        visited: Set[int] = set()

        def recurse(obj: Any) -> None:
            oid = id(obj)
            if oid in visited:
                return
            visited.add(oid)

            # 1) Handle TTLayer instances (leaf components)
            if isinstance(obj, TTLayer):
                fn = getattr(obj, method_name, None)
                if callable(fn):
                    try:
                        fn(*args, **kwargs)
                    except TypeError:
                        fn(*args, **kwargs)

            # 2) Handle TTModel instances that override the method
            elif isinstance(obj, TTModel):
                cls_fn = getattr(obj.__class__, method_name, None)
                base_fn = getattr(TTModel, method_name, None)
                if cls_fn is not None and cls_fn is not base_fn:
                    try:
                        cls_fn(obj)
                    except TypeError:
                        cls_fn(obj)

            # 3) Handle other objects that have the method
            elif hasattr(obj, method_name):
                fn = getattr(obj, method_name, None)
                if callable(fn):
                    try:
                        fn()
                    except TypeError:
                        fn()

            # 4) Recurse into registered submodules (this covers ModuleList, Sequential, etc.)
            if isinstance(obj, nn.Module):
                for child in obj._modules.values():
                    if child is None:
                        continue
                    recurse(child)

            # 5) Recurse into container attributes that might hold modules or other objects with methods.
            #    This covers plain python lists/tuples/dicts/sets assigned as attributes on modules.
            #    We purposely ignore common atomic types (tensors, numbers, strings).
            try:
                attrs = getattr(obj, "__dict__", {})
            except Exception:
                attrs = {}

            for attr in attrs.values():
                if attr is None:
                    continue
                # common containers
                if isinstance(attr, (list, tuple, set)):
                    for el in attr:
                        recurse(el)
                elif isinstance(attr, dict):
                    for el in attr.values():
                        recurse(el)
                else:
                    # If attribute is a module or object, recurse (visited prevents duplicates)
                    if isinstance(attr, (nn.Module, object)):
                        recurse(attr)

        # start recursion from self
        recurse(self)

    def TTcompile(self):
        """Recursively compile all TTLayer instances for inference"""
        self._call_recursive("TTcompile")

    def TTuncompile(self):
        """Recursively uncompile all TTLayer instances"""
        self._call_recursive("TTuncompile")



from typing import TypedDict, Optional, Literal, Union, Dict, Any
from ._leaky_integrator import LeakyIntegrator
from ... import functional
import torch
from torch import nn


class LIT(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            beta: Union[float, torch.Tensor] = 0.9,
            pos_threshold: Union[float, torch.Tensor] = 1.0,
            neg_threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            beta_rank: Literal[0, 1] = 1,
            pos_threshold_rank: Literal[0, 1] = 1,
            neg_threshold_rank: Literal[0, 1] = 1,
            learn_beta: bool = True,
            learn_pos_threshold: bool = True,
            learn_neg_threshold: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass


class DLIT(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            pos_threshold: Union[float, torch.Tensor] = 1.0,
            neg_threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            pos_threshold_rank: Literal[0, 1] = 1,
            neg_threshold_rank: Literal[0, 1] = 1,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
            learn_pos_threshold: bool = True,
            learn_neg_threshold: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass


class SLIT(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            alpha: Union[float, torch.Tensor] = 0.5,
            beta: Union[float, torch.Tensor] = 0.9,
            pos_threshold: Union[float, torch.Tensor] = 1.0,
            neg_threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            alpha_rank: Literal[0, 1] = 1,
            beta_rank: Literal[0, 1] = 1,
            pos_threshold_rank: Literal[0, 1] = 1,
            neg_threshold_rank: Literal[0, 1] = 1,
            learn_alpha: bool = True,
            learn_beta: bool = True,
            learn_pos_threshold: bool = True,
            learn_neg_threshold: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass


class RLIT(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            beta: Union[float, torch.Tensor] = 0.9,
            gamma: Union[float, torch.Tensor] = 0.9,
            pos_threshold: Union[float, torch.Tensor] = 1.0,
            neg_threshold: Union[float, torch.Tensor] = 1.0,
            rec_weight: Union[float, torch.Tensor] = 0.0,
            dim: int = -1,
            beta_rank: Literal[0, 1] = 1,
            gamma_rank: Literal[0, 1] = 1,
            pos_threshold_rank: Literal[0, 1] = 1,
            neg_threshold_rank: Literal[0, 1] = 1,
            rec_weight_rank: Literal[0, 1] = 1,
            learn_beta: bool = True,
            learn_gamma: bool = True,
            learn_pos_threshold: bool = True,
            learn_neg_threshold: bool = True,
            learn_rec_weight: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass


class DSLIT(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            pos_alpha: Union[float, torch.Tensor] = 0.5,
            neg_alpha: Union[float, torch.Tensor] = 0.5,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            pos_threshold: Union[float, torch.Tensor] = 1.0,
            neg_threshold: Union[float, torch.Tensor] = 1.0,
            dim: int = -1,
            pos_alpha_rank: Literal[0, 1] = 1,
            neg_alpha_rank: Literal[0, 1] = 1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            pos_threshold_rank: Literal[0, 1] = 1,
            neg_threshold_rank: Literal[0, 1] = 1,
            learn_pos_alpha: bool = True,
            learn_neg_alpha: bool = True,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
            learn_pos_threshold: bool = True,
            learn_neg_threshold: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass


class DRLIT(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            pos_gamma: Union[float, torch.Tensor] = 0.9,
            neg_gamma: Union[float, torch.Tensor] = 0.9,
            pos_threshold: Union[float, torch.Tensor] = 1.0,
            neg_threshold: Union[float, torch.Tensor] = 1.0,
            pos_rec_weight: Union[float, torch.Tensor] = 0.0,
            neg_rec_weight: Union[float, torch.Tensor] = 0.0,
            dim: int = -1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            pos_gamma_rank: Literal[0, 1] = 1,
            neg_gamma_rank: Literal[0, 1] = 1,
            pos_threshold_rank: Literal[0, 1] = 1,
            neg_threshold_rank: Literal[0, 1] = 1,
            pos_rec_weight_rank: Literal[0, 1] = 1,
            neg_rec_weight_rank: Literal[0, 1] = 1,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
            learn_pos_gamma: bool = True,
            learn_neg_gamma: bool = True,
            learn_pos_threshold: bool = True,
            learn_neg_threshold: bool = True,
            learn_pos_rec_weight: bool = True,
            learn_neg_rec_weight: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass


class SRLIT(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            alpha: Union[float, torch.Tensor] = 0.5,
            beta: Union[float, torch.Tensor] = 0.9,
            gamma: Union[float, torch.Tensor] = 0.9,
            pos_threshold: Union[float, torch.Tensor] = 1.0,
            neg_threshold: Union[float, torch.Tensor] = 1.0,
            rec_weight: Union[float, torch.Tensor] = 0.0,
            dim: int = -1,
            alpha_rank: Literal[0, 1] = 1,
            beta_rank: Literal[0, 1] = 1,
            gamma_rank: Literal[0, 1] = 1,
            pos_threshold_rank: Literal[0, 1] = 1,
            neg_threshold_rank: Literal[0, 1] = 1,
            rec_weight_rank: Literal[0, 1] = 1,
            learn_alpha: bool = True,
            learn_beta: bool = True,
            learn_gamma: bool = True,
            learn_pos_threshold: bool = True,
            learn_neg_threshold: bool = True,
            learn_rec_weight: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass


class DSRLIT(LeakyIntegrator):
    def __init__(
            self,
            num_neurons: int,
            pos_alpha: Union[float, torch.Tensor] = 0.5,
            neg_alpha: Union[float, torch.Tensor] = 0.5,
            pos_beta: Union[float, torch.Tensor] = 0.9,
            neg_beta: Union[float, torch.Tensor] = 0.9,
            pos_gamma: Union[float, torch.Tensor] = 0.9,
            neg_gamma: Union[float, torch.Tensor] = 0.9,
            pos_threshold: Union[float, torch.Tensor] = 1.0,
            neg_threshold: Union[float, torch.Tensor] = 1.0,
            pos_rec_weight: Union[float, torch.Tensor] = 0.0,
            neg_rec_weight: Union[float, torch.Tensor] = 0.0,
            dim: int = -1,
            pos_alpha_rank: Literal[0, 1] = 1,
            neg_alpha_rank: Literal[0, 1] = 1,
            pos_beta_rank: Literal[0, 1] = 1,
            neg_beta_rank: Literal[0, 1] = 1,
            pos_gamma_rank: Literal[0, 1] = 1,
            neg_gamma_rank: Literal[0, 1] = 1,
            pos_threshold_rank: Literal[0, 1] = 1,
            neg_threshold_rank: Literal[0, 1] = 1,
            pos_rec_weight_rank: Literal[0, 1] = 1,
            neg_rec_weight_rank: Literal[0, 1] = 1,
            learn_pos_alpha: bool = True,
            learn_neg_alpha: bool = True,
            learn_pos_beta: bool = True,
            learn_neg_beta: bool = True,
            learn_pos_gamma: bool = True,
            learn_neg_gamma: bool = True,
            learn_pos_threshold: bool = True,
            learn_neg_threshold: bool = True,
            learn_pos_rec_weight: bool = True,
            learn_neg_rec_weight: bool = True,
            surrogate_derivative=functional.atan_surrogate(2.0),
    ):
        pass



import torch
from .. import plot


class MeasurementManager:
	"""
	Manages measurements over time like loss or accuracy
	"""

	def __init__(
			self,
			title: str,
			decay: list = [0., 0.9, 0.99, 0.999]
	):
		self.title = title
		self.decay = torch.tensor(decay).float()
		self.trace = torch.zeros_like(self.decay)
		self.measurement = []

	def append(self, value):
		with torch.no_grad():
			if isinstance(value, torch.Tensor):
				value = value.item()
			self.trace *= self.decay
			self.trace += value * torch.ones_like(self.trace)
			avg_input = self.trace * (1 - self.decay)
			self.measurement.append(avg_input)

	def plot(self, title: str = None):
		plot_title = title if title is not None else self.title
		rounded_list = [round(x, 4) for x in self.decay.tolist()]
		plot.line_graph(self.measurement, title=plot_title, label=rounded_list)



from ._decay_conversion import halflife_to_decay, decay_to_halflife, timesteps_to_decay, decay_to_timesteps
from ._inverse_functions import sigmoid_inverse, softplus_inverse
from ._spike_functions import sigmoid4x
from ._surrogates import round_ste, bernoulli_ste, probabilistic_ste



import torch


def soft_cross_entropy(pred: torch.Tensor, target: torch.Tensor, eps: float = 1e-8, reduction: str = 'mean'):
	"""
	Generalized soft cross entropy between two probability tensors of the same shape.

	pred:   [B, ...] predicted probabilities (e.g., from softmax)
	target: [B, ...] target probabilities (e.g., one-hot, Dirichlet samples, etc.)
	eps:    Small value for numerical stability
	reduction: 'none' | 'mean' | 'sum'
	"""
	if pred.shape != target.shape:
		raise ValueError(f"Shape mismatch: pred {pred.shape} vs target {target.shape}")

	pred = torch.clamp(pred, eps, 1.0)
	loss = -(target * torch.log(pred))

	# Sum over all non-batch dims
	loss = loss.view(loss.size(0), -1).sum(dim=1)

	if reduction == 'mean':
		return loss.mean()
	elif reduction == 'sum':
		return loss.sum()
	elif reduction == 'none':
		return loss
	else:
		raise ValueError(f"Invalid reduction: {reduction}")
